# coding: utf8

"""
scripts to handle mturk commands, using Drupal Hybrid Computing module and the MTurk Java SDK at:
http://aws.amazon.com/code/Amazon-Mechanical-Turk/695
Mturk's Java SDK has more functionality than the Python Boto version.

requires jython 2.7+
"""

import sys, time, json
from pprint import pprint
from collections import defaultdict

from org.drupal.project.computing.common import PingMe
from org.drupal.project.computing.jython import DJyApplication, DJyCommand
from org.drupal.project.computing import DUtils, DCommand, DCommandExecutionException
from com.amazonaws.mturk.util import ClientConfig
from com.amazonaws.mturk.service.axis import RequesterService
from com.amazonaws.mturk.addon import HITProperties
from com.amazonaws.mturk.service.exception import InternalServiceException, ServiceException


cv = lambda v: v.getValue() if v is not None else None
ct = lambda t: int(t.getTimeInMillis()/1000) if t is not None else None


class MTurkApplication(DJyApplication):

  # Override
  def getIdentifier(self):
    return 'mturk'

  # Override
  def buildCommandRegistry(self):
    self.registerCommand(PingMe)
    # data analysis command
    self.registerCommand('test', MTTestCommand)
    self.registerCommand('DataAnalysis', DataAnalysis)

    # service command
    self.registerCommand('GetAccountBalance', GetAccountBalance)
    self.registerCommand('CreateHITs', CreateHITs)
    self.registerCommand('RegisterHITType', RegisterHITType)
    self.registerCommand('LoadAllHITs', LoadAllHITs)
    self.registerCommand('LoadAllAssignments', LoadAllAssignments)
    self.registerCommand('LoadCreatedHITs', LoadCreatedHITs)
    self.registerCommand('RemoveCreatedHITs', RemoveCreatedHITs)
    self.registerCommand('GrantBonus', GrantBonus)
    self.registerCommand('SendMessage', SendMessage)
    self.registerCommand('RetrieveAllQualificationTypes', RetrieveAllQualificationTypes)
    self.registerCommand('QualificationTypeInfo', QualificationTypeInfo)
    self.registerCommand('ApproveLoadedAssignments', ApproveLoadedAssignments)
    self.registerCommand('ExtendHITs', ExtendHITs)



##########################################################

class MTurkCommand(DJyCommand):

  # Overrides
  def getIdentifier(self):
    # by default, just use the class name.
    return self.__class__.__name__


  # Abstract method, return Status/Messages
  #def execute(self):
  #  raise NotImplementedError('Pleaes override execute() method or the run() method.')


  # Overrides
  def enterRecord(self, record):
    inputjson = record.getInputjson()
    input = json.loads(inputjson) if inputjson is not None else {}
    self.initialize(record.getUid(), record.getNid(), input)


  def mapArgs(self, record, args):
    assert len(args) in [2, 3], 'Args: ' + str(args)
    record.setUid(int(args[0]))
    record.setNid(int(args[1]))
    if (len(args) == 3):
      record.setInputjson(args[2])
      

  def keepRecord(self, record):
    if hasattr(self, 'output'):
      outputjson = json.dumps(self.output)
      record.setOutputjson(outputjson)


  def initialize(self, user_id, task_id, input = {}):
    self.logger = DUtils.getInstance().getPackageLogger()
    self.user_id, self.task_id = user_id, task_id
    self.input = input

    self.timestamp = int(time.time())
    self.drush = self.getDrupalSite().getDrush()
    if not hasattr(self, 'settings'):
      self.settings = {}

    # read user settings
    if self.user_id is not None:
      user = json.loads(self.drush.computingEval('return user_load(' + str(self.user_id) + ');'))
      if len(user['field_mturk_settings']) > 0:
        self.settings.update(DUtils.getInstance().loadProperties(user['field_mturk_settings']['und'][0]['safe_value']))

    # read task settings, will override the user's settings if duplicate.
    if self.task_id is not None:
      node = json.loads(self.drush.computingEval('return node_load(' + str(self.task_id) + ');'))
      if len(node['field_mturk_settings']) > 0:
        self.original_task_settings = node['field_mturk_settings']['und'][0]['value']
        self.settings.update(DUtils.getInstance().loadProperties(self.original_task_settings))



class MTurkServiceCommand(MTurkCommand):
  """ Sub class for any mturk command related to the SDK."""

  # this is hard coded and default to all mturk_task instances.
  # TODO: could be from mturk.properties.
  def default_service_config(self):
    config = ClientConfig()
    config.setRetriableErrors(set(['503', 'AWS.ServiceUnavailable', 'Server.ServiceUnavailable']))
    config.setRetryAttempts(10)
    config.setRetryDelayMillis(1000)
    return config



  def initialize(self, user_id, task_id, input):
    super(MTurkServiceCommand, self).initialize(user_id, task_id, input)
    assert self.user_id != None and self.task_id != None, 'MTurk UserID and TaskID cannot be empty.'

    access_key, secret_key = self.settings['access_key'], self.settings['secret_key']
    assert access_key != None and secret_key != None, 'MTurk "access key" and "secret key" cannot be empty.'

    # config service
    config = self.default_service_config()
    if self.settings.has_key('use_sandbox') and self.settings['use_sandbox'].lower() in ['true', '1', 'yes', 'y']:
      config.setServiceURL(ClientConfig.SANDBOX_SERVICE_URL)
      self.WORKER_WEBSITE_URL = ClientConfig.SANDBOX_WORKER_WEBSITE_URL
    else:
      config.setServiceURL(ClientConfig.PRODUCTION_SERVICE_URL)
      self.WORKER_WEBSITE_URL = ClientConfig.PROD_WORKER_WEBSITE_URL
    config.setAccessKeyId(access_key)
    config.setSecretAccessKey(secret_key)

    # connect to service
    self.service = RequesterService(config)



class MTurkDataCommand(MTurkCommand):
  """Super class for all commands that don't require mturk SDK, but will do work with MTurk data"""
  pass


########################################################

class MTTestCommand(MTurkDataCommand):

  def getIdentifier(self):
    return 'test'

  def execute(self):
    self.printMessage('Execution successful.')



class GetAccountBalance(MTurkServiceCommand):

  def keepRecord(self, record):
    super(GetAccountBalance, self).keepRecord(record)
    record.setNumber2(self.balance)

  def execute(self):
    self.balance = self.service.getAccountBalance()
    msg = "Account balance: " + str(self.balance)
    self.logger.info(msg)
    self.printMessage(msg)
    self.output = {'balance' : self.balance}



EXTERNAL_QUESTION_TEMPLATE = '''
<ExternalQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd">
  <ExternalURL>%s</ExternalURL>
  <FrameHeight>%d</FrameHeight>
</ExternalQuestion>
'''

class CreateHITs(MTurkServiceCommand):

  def execute(self):
    items = self.input

    register_hittype = DCommand.create(RegisterHITType, self.getApplication())
    register_hittype.mapParams(register_hittype.getRecord(), self.user_id, self.task_id)
    register_hittype.run()
    hittypeid = register_hittype.hittypeid
    self.logger.info('Create HIT with HIT TYPE ID: ' + hittypeid)
    
    default_frame_height = int(self.settings.get('frameheight', 400))
    default_hit_life_time = int(self.settings.get('hitlifetime', 60*60*24))  # one day.
    default_max_assignments = int(self.settings.get('assignments', 1))
    self.hits, hits_id = [], []

    for id, value in items.items():
      url = value['url']
      frame_height = value.get('frame_height', default_frame_height)
      hit_life_time = DUtils.getInstance().getLong(value.get('hit_life_time', default_hit_life_time))
      max_assignments = DUtils.getInstance().getLong(value.get('max_assignments', default_max_assignments))
      question = EXTERNAL_QUESTION_TEMPLATE % (url, frame_height)

      try:
      # call the service and create hit.
        hit = self.service.createHIT(
          hittypeid,
          None, # title
          None, # description
          None, # keywords
          question,    # we have question here, so can't use the HITLayout API.
          None, # reward
          None, # assignmentDurationInSeconds
          None, # autoApprovalDelayInSeconds
          hit_life_time,
          max_assignments, # maxAssignments
          id,   # annotation
          None, # qualificationRequirements
          ['Minimal'], # responseGroup, only need the HIT id
          "%d|%s" % (self.task_id, id),   # uniqueRequestToken
          None, # assignmentReviewPolicy
          None, # hitReviewPolicy
        )
        self.hits.append(hit)
        hits_id.append(hit.getHITId())
        self.logger.finest('Created HIT: ' + hit.getHITId())
      except InternalServiceException as e:
        error_message = e.getMessage()
        if error_message.find('HITAlreadyExists') > 0:
          hit_id = error_message.split('"')[1]
          hits_id.append(hit_id)
          self.logger.finest('HIT already exists: %s' % hit_id)
        else:
          self.logger.warning('Mturk Service Error: %s' % error_message)
          e.printStackTrace()

    self.output = {
      'hittypeid': hittypeid,
      'created':   hits_id,
    }
    self.printMessage('Created %d HITs with HIT type id: %s' % (len(hits_id), hittypeid))

  def keepRecord(self, record):
    super(CreateHITs, self).keepRecord(record)
    record.setString2(self.WORKER_WEBSITE_URL + '/mturk/preview?groupId=' + self.output['hittypeid'])


class ExtendHITs(MTurkServiceCommand):
  def enterRecord(self, record):
    self.max_assignment_increment = int(record.getNumber1()) if record.getNumber1() is not None and int(record.getNumber1()) > 0 else None
    self.expiration_increment = int(record.getNumber2()) if record.getNumber2() is not None and int(record.getNumber2()) > 0 else None
    #assert self.max_assignment_increment >= 0 and self.expiration_increment >= 0
    super(ExtendHITs, self).enterRecord(record)
  
  def mapArgs(self, record, args):
    assert len(args) == 4, 'Args: ' + str(args)
    record.setUid(int(args[0])) # user_id
    record.setNid(int(args[1])) # task_id
    record.setNumber1(int(args[2])) # max_assignment_increment
    record.setNumber2(int(args[3]))  # expiration_increment.
    
  def execute(self):
    self.db = self.getDrupalSite().getDatabase()
    hits = self.db.queryArray("SELECT hit_id FROM {mturk_task_hit} WHERE task_id=?", self.task_id)
    print "Extends for", len(hits), "HITs. Please wait."
    for hit_id, in hits:
      self.service.extendHIT(hit_id, self.max_assignment_increment, self.expiration_increment)
    self.printMessage('Successfully extended HITs.')
    self.db.close()
    
    


class RegisterHITType(MTurkServiceCommand):
  def mapParams(self, record, user_id, task_id):
    assert user_id is not None and task_id is not None
    record.setUid(user_id)
    record.setNid(task_id)
  
  def execute(self):
    assert hasattr(self, 'original_task_settings')
    props = HITProperties(DUtils.getInstance().loadProperties(self.original_task_settings))
    self.hittypeid = self.service.registerHITType(
      props.getAutoApprovalDelay(),
      props.getAssignmentDuration(),
      props.getRewardAmount(),
      props.getTitle(),
      props.getKeywords(),
      props.getDescription(),
      props.getQualificationRequirements()
    )
    self.printMessage('HITType registered: ' + self.hittypeid)
    self.output = {'hittypeid': self.hittypeid}
    # TODO: save to mturk_hittype as well.



class LoadAllHITs(MTurkServiceCommand):

  def execute(self):
    self.db = self.getDrupalSite().getDatabase();

    print "Retrieving all HITs. Please wait. Task:", self.task_id
    hits = self.service.searchAllHITs()
    print 'Total number of HITs:', len(hits)

    print 'Now saving to database. WARNING: old HITs will be deleted'
    deleted = self.db.update('DELETE FROM {mturk_task_hit} WHERE task_id=?', self.task_id)
    print 'Deleted old HITs:', deleted

    # TODO: allow people to assign hit_type_id and only receive those assigned.
    # current approach is that after loadAllHits, run LoadHits, and will remove all hit_type_ids not specified.

    # process hit types and save to db.
    hit_types = defaultdict(int)
    for hit in hits:
      hit_types[hit.getHITTypeId()] += 1
      self.save_hit_to_db(hit)
    print 'Total number of HITTypes:', len(hit_types)
    self.printMessage('Total # of HITs loaded: '+str(len(hits))+' -- HIT Types and # of HITs: '+'; '.join([k+':'+str(v) for k,v in hit_types.items()]))
    self.db.close()


  # save a hit to db
  def save_hit_to_db(self, hit):
    #print 'Processing:', hit.getHITId()
    # save hit_type
    if self.db.queryValue("SELECT hit_type_id FROM {mturk_hit_type} WHERE hit_type_id=?", hit.getHITTypeId()) == None:
      # insert hit_type
      self.db.update("INSERT INTO {mturk_hit_type} (hit_type_id, title, reward_amount, reward_currency, assignment_duration, auto_approval_delay, description, keywords, updated) \
                  VALUE (?, ?, ?, ?, ?, ?, ?, ?, ?)", hit.getHITTypeId(), hit.getTitle(), hit.getReward().getAmount(), hit.getReward().getCurrencyCode(),
        hit.getAssignmentDurationInSeconds(), hit.getAutoApprovalDelayInSeconds(), hit.getDescription(), hit.getKeywords(), self.timestamp)
    else:
      # update changes
      self.db.update("UPDATE {mturk_hit_type} SET title=?, reward_amount=?, reward_currency=?, assignment_duration=?, auto_approval_delay=?, description=?, keywords=?, updated=? \
                  WHERE hit_type_id=?", hit.getTitle(), hit.getReward().getAmount(), hit.getReward().getCurrencyCode(),
        hit.getAssignmentDurationInSeconds(), hit.getAutoApprovalDelayInSeconds(), hit.getDescription(), hit.getKeywords(), self.timestamp, hit.getHITTypeId())

    # save hit
    if self.db.queryValue("SELECT hit_id FROM {mturk_hit} WHERE hit_id=?", hit.getHITId()) == None:
      # insert hit
      self.db.update("INSERT INTO {mturk_hit} (hit_id, hit_type_id, annotation_id, question, hit_review_status, hit_status, max_assignments, hit_creation, hit_expiration, updated) \
                  VALUE (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", hit.getHITId(), hit.getHITTypeId(), hit.getRequesterAnnotation(), hit.getQuestion(), cv(hit.getHITReviewStatus()),
        cv(hit.getHITStatus()), hit.getMaxAssignments(), ct(hit.getCreationTime()), ct(hit.getExpiration()), self.timestamp)
    else:
      # update hit
      self.db.update("UPDATE {mturk_hit} SET hit_type_id=?, annotation_id=?, question=?, hit_review_status=?, hit_status=?, max_assignments=?, hit_expiration=?, updated=? \
                  WHERE hit_id=?", hit.getHITTypeId(), hit.getRequesterAnnotation(), hit.getQuestion(), cv(hit.getHITReviewStatus()),
        cv(hit.getHITStatus()), hit.getMaxAssignments(), ct(hit.getExpiration()), self.timestamp, hit.getHITId())

    if self.db.queryValue("SELECT 1 FROM {mturk_task_hit} WHERE task_id = ? AND hit_id = ?", self.task_id, hit.getHITId()) == None:
      self.db.update("INSERT INTO {mturk_task_hit} (task_id, hit_id, updated) VALUE (?, ?, ?)", self.task_id, hit.getHITId(), self.timestamp)
    else:
      # duplicate data, we don't do anything.
      pass



class LoadCreatedHITs(LoadAllHITs):
  """
  Regardless of the status of the HIT, we'll load them. See GetReviewableHITs for another example.
  """

  def execute(self):
    self.db = self.getDrupalSite().getDatabase();
    print "Retrieving created HITs. Task:", self.task_id

    hits = []
    rows = self.db.query("SELECT outputjson FROM {computing_record} WHERE nid=? AND command='CreateHITs' AND status='OKOK'", self.task_id)
    for row in rows:
      output = json.loads(row['outputjson'])
      hits += output['created']
    hits = set(hits)  # dedup
    print 'Total number of HITs:', len(hits)

    # process hit and save to db.
    for hit_id in hits:
      hit = self.service.getHIT(hit_id)
      self.save_hit_to_db(hit)
    self.printMessage('Total # of HITs loaded: '+str(len(hits)))
    self.db.close()


# TODO: test this. haven't tested yet.
class GetReviewableHITs(LoadAllHITs):

  def execute(self):
    self.db = self.getDrupalSite().getDatabase();
    #self.logger.info("Retrieving reviewable HITs. Task: %d" % self.task_id)

    hit_types = []
    rows = self.db.query("SELECT outputjson FROM {computing_record} WHERE nid=? AND command='CreateHITs'", self.task_id)
    for row in rows:
      output = json.loads(row['outputjson'])
      hit_types += output['hittypeid']
    hit_types = set(hit_types)  # dedup

    reviewable_hits = []
    for hit_type in hit_types:
      self.logger.info("Retrieving reviewalbe HITS for HIT Type %s" % hit_type)
      reviewable_hits += self.service.getAllReviewableHITs(hit_type)

    for hit in reviewable_hits:
      self.save_hit_to_db(hit)

    self.printMessage('Total # of reviewable HITs loaded: '+str(len(reviewable_hits)))
    self.db.close()



class LoadAllAssignments(MTurkServiceCommand):

  def execute(self):
    self.db = self.getDrupalSite().getDatabase()
    hits = self.db.queryArray("SELECT hit_id FROM {mturk_task_hit} WHERE task_id=?", self.task_id)
    print "Loading assignments for", len(hits), "HITs. Please wait."
    total = 0
    for hit_id, in hits:
      assignments = self.service.getAllAssignmentsForHIT(hit_id)
      for assignment in assignments:
        self.save_assignment_to_db(assignment)
        total += 1
    self.printMessage('Total assignments loaded: %d' % total)
    self.db.close()


  def save_assignment_to_db(self, assignment):
    assignment_status = assignment.getAssignmentStatus()
    if assignment_status is not None:
      assignment_status = assignment_status.getValue()

    assignment_id = assignment.getAssignmentId()
    answer_xml = assignment.getAnswer()
    answer_json = self.parse_answer_xml(assignment_id, answer_xml)

    if self.db.queryValue("SELECT assignment_id FROM {mturk_assignment} WHERE assignment_id=?", assignment.getAssignmentId()) == None:
      # insert new assignment
      self.db.update("INSERT {mturk_assignment} (assignment_id, hit_id, worker_id, assignment_status, accept_time, submit_time, \
                  approval_time, rejection_time, answer_xml, answer_json, requester_feedback, updated) VALUE (?,?,?,?,?,?,?,?,?,?,?,?)",
        assignment_id, assignment.getHITId(), assignment.getWorkerId(), assignment_status, ct(assignment.getAcceptTime()),
        ct(assignment.getSubmitTime()), ct(assignment.getApprovalTime()), ct(assignment.getRejectionTime()),
        answer_xml, answer_json, assignment.getRequesterFeedback(), self.timestamp)
    else:
      # update existing assignment
      self.db.update("UPDATE {mturk_assignment} SET assignment_status=?, accept_time=?, submit_time=?, approval_time=?, rejection_time=?,\
                  answer_xml=?, answer_json=?, requester_feedback=?, updated=? WHERE assignment_id=?",
        assignment_status, ct(assignment.getAcceptTime()),
        ct(assignment.getSubmitTime()), ct(assignment.getApprovalTime()), ct(assignment.getRejectionTime()),
        answer_xml, answer_json, assignment.getRequesterFeedback(), self.timestamp, assignment_id)


  def parse_answer_xml(self, assignment_id, answer_xml):
    ''' Return json string '''
    answers_json = {}
    answers = self.service.parseAnswers(answer_xml)
    for answer in answers.getAnswer():
      answer_str = self.service.getAnswerValue(assignment_id, answer, True)
      # according to the implementation of RequesterService.getAnswerValue(), '\t' is the delimeter for question and answer strings.
      delim_pos = answer_str.find('\t')
      q, a = answer_str[:delim_pos], answer_str[delim_pos+1:]
      answers_json[q] = a
    return json.dumps(answers_json)
  
  

class BatchGrantBonus(MTurkServiceCommand):
  """
  The caller is responsible to prepare assignment_id, but can grant bonus to multiple workers.
  """

  def execute(self):
    failed = []
    piece, total = 0, 0
    for worker_id, bonus_amount, assignment_id in self.input['grant']:
      try:
        self.service.grantBonus(worker_id, bonus_amount, assignment_id, self.input['message'])
        self.logger.info('Granted bonus %3.2f to worker %s' % (bonus_amount, worker_id))
        piece += 1
        total += bonus_amount
      except:
        failed.append([worker_id, bonus_amount, assignment_id])
        self.logger.warning('Cannot grant bonus: %s, %3.2f, %s' % (worker_id, bonus_amount, assignment_id))
    self.output['failed'] = failed
    self.printMessage("Successfully made %d grants with a total of %3.2f." % (piece, total))



class GrantBonus(MTurkServiceCommand):
 
  def enterRecord(self, record):
    self.worker_id = record.getString1()
    self.reason = record.getString2()
    # assignment_id is optional. if not present, then select the last assignment in task_id.
    if record.getString3() is not None:
      self.assignment_id = record.getString3()
    self.bonus_amount = record.getNumber1()
    super(GrantBonus, self).enterRecord(record)
  
  
  def mapArgs(self, record, args):
    assert len(args) == 5, 'Args: ' + str(args)
    record.setUid(int(args[0])) # user_id
    record.setNid(int(args[1])) # task_id
    record.setString1(args[2]) # worker_id
    record.setNumber1(float(args[3])) # bonus amount
    record.setString2(args[4])  # reason message
    
    
    
  def execute(self):
    if not hasattr(self, 'assignment_id'):
      drush = self.getDrupalSite().getDrush()
      json_results = drush.computingCall(['mturk_bonus_get_assignment_by_worker', json.dumps(self.worker_id), json.dumps(self.task_id)])
      #pprint(json_results)
      self.assignment_id = json.loads(json_results)
      #pprint(self.assignment_id)
    if self.assignment_id is None or self.assignment_id == False:
      raise DCommandExecutionException('Invalid assignment ID')
    
    self.service.grantBonus(self.worker_id, self.bonus_amount, self.assignment_id, self.reason)
    self.printMessage('Granted bonus %3.2f to worker %s' % (self.bonus_amount, self.worker_id))



class SendMessage(MTurkServiceCommand):
  MESSAGE_BATCH_SIZE = 50 # 100 is the upper limit for one batch.
  
  def enterRecord(self, record):
    self.subject = record.getString1()
    self.body = record.getString2()
    worker_string = record.getString3() 
    self.workers = worker_string.split('|')
    assert len(self.subject) > 0 and len(self.body) > 0 and len(self.workers) > 0
    #pprint(self.subject)
    #pprint(self.body)
    #pprint(self.workers)
    super(SendMessage, self).enterRecord(record)
  
  def mapArgs(self, record, args):
    assert len(args) == 5, 'Args: ' + str(args)
    record.setUid(int(args[0])) # user_id
    record.setNid(int(args[1])) # task_id
    record.setString1(args[2]) # subject
    record.setString2(args[3]) # message body
    record.setString3(args[4]) # a list of worker ids

  def execute(self):
    batch_size = self.MESSAGE_BATCH_SIZE
    batches = (len(self.workers)-1)/batch_size + 1
    for batch in range(batches):
      rp = self.workers[batch*batch_size : (batch+1)*batch_size]
      #print "Sending messages to", len(rp), "recipients"
      self.service.notifyWorkers(self.subject, self.body, rp)
      self.logger.fine('Sent message to %d workers' % len(rp))
    self.printMessage('Sent message to a total of %d workers' % len(self.workers))
    
    
class QualificationTypeInfo(MTurkServiceCommand):
  
  def enterRecord(self, record):
    self.qualification_type_id = record.getString1()
    super(QualificationTypeInfo, self).enterRecord(record)
    
  def mapArgs(self, record, args):
    assert len(args) == 3, 'Args: ' + str(args)
    record.setUid(int(args[0])) # user_id
    record.setNid(int(args[1])) # task_id
    record.setString1(args[2]) # qualification id.
    
  def execute(self):
    # list qualification type info
    qualification_type = self.service.getQualificationType(self.qualification_type_id)
    print(DUtils.getInstance().objectToString(qualification_type))
    
    # list qualification requests
    #qualification_requests = self.service.getAllQualificationRequests(self.qualification_type_id)
    #for r in qualification_requests:
    #  print(DUtils.getInstance().objectToString(r))
    
    # list qualification grants/scores
    qualifications = self.service.getAllQualificationsForQualificationType(self.qualification_type_id)
    for q in qualifications:
      print(DUtils.getInstance().objectToString(r))
      
    #self.printMessage('Print qualification type info successful. See execution log for details. Total requests: %d. Total granted: %d' % \
    #                  (len(qualification_requests), len(qualifications)))
    self.printMessage('Print qualification type info successful. See execution log for details. Total granted: %d' %  len(qualifications))
    
    

class RetrieveAllQualificationTypes(MTurkServiceCommand):
  def execute(self):
    self.qualifications = self.service.getAllQualificationTypes()
    self.printMessage('Retrieved qualification types: %d' % len(self.qualifications))
    
  def keepRecord(self, record):
    ids = []
    for q in self.qualifications:
      ids.append(q.getQualificationTypeId())
    record.setString1(','.join(ids))
      
    
class RemoveCreatedHITs(MTurkServiceCommand):
  def execute(self):
    self.db = self.getDrupalSite().getDatabase()
    print "Retrieving created HITs. Task:", self.task_id

    hits = []
    rows = self.db.query("SELECT outputjson FROM {computing_record} WHERE nid=? AND command='CreateHITs' AND status IS NOT NULL", self.task_id)
    for row in rows:
      output = json.loads(row['outputjson'])
      hits += output['created']
    hits = set(hits)  # dedup
    print 'Total number of HITs:', len(hits)

    # process hit and save to db.
    removed = 0
    for hit_id in hits:
      try:
        hit = self.service.disableHIT(hit_id)
        removed += 1
      except ServiceException as e:
        self.logger.warning('Cannot remove HIT. Could be removed already: ' + hit_id)
    self.printMessage('Total # of HITs disabled: %d' % removed)
    self.db.close()


class ApproveLoadedAssignments(MTurkServiceCommand):
  def enterRecord(self, record):
    if record.getString1() is not None:
      self.feedback = record.getString1()
    else:
      self.feedback = None
    super(ApproveLoadedAssignments, self).enterRecord(record)

  def execute(self):
    db = self.getDrupalSite().getDatabase()
    assignments = []
    rows = db.query("SELECT assignment_id FROM {mturk_assignment} a INNER JOIN {mturk_task_hit} t "
                    "ON a.hit_id=t.hit_id WHERE t.task_id=? AND assignment_status='Submitted'", self.task_id)
    self.logger.info('Assignments to be approved: %d' % len(rows))
    db.close()

    success = 0
    for row in rows:
      try:
        self.service.approveAssignment(row['assignment_id'], self.feedback)
        success += 1
      except ServiceException as e:
        self.logger.warning('Cannot approave assignment %s. Might be approved already.' % row['assignment_id'])

    self.printMessage('Assignment approved: %d. Reload assignments to see status change.' % success)




class DataAnalysis(MTurkDataCommand):
  
  def _prepare_first(self):
    assignments = self.db.query("SELECT t.hit_id, a.worker_id, a.assignment_status, a.answer_json, a.accept_time, a.submit_time\
        FROM {mturk_task_hit} t INNER JOIN {mturk_assignment} a ON t.hit_id=a.hit_id AND task_id=? ORDER BY t.hit_id", self.task_id)
    if not len(assignments):
      self.printMessage('Total assignments is 0, please load assignments first before computing results.')
      raise DCommandExecutionException("Total assignments is 0, please load assignments first before computing results.")
    
    self.assignments = assignments

    answers_summary = {}  # number of answers for each answer (needs to specify the 'main question' setting)
    hit_group = defaultdict(dict)  # group all assignments according to hit_id. if not assignments, then no hit_id as key.
    hit_group_info = defaultdict(list) # group all assignment according to hit_id
    
    for row in assignments:
      hit_id, worker_id, assignment_status, answer_json, accept_time, submit_time = row['hit_id'], row['worker_id'], row['assignment_status'], row['answer_json'], row['accept_time'], row['submit_time']
      answer_json = json.loads(answer_json)
      answer = answer_json[self.main_question]  # here, we assume each assignments need to answer the main question.
      hit_group[hit_id][worker_id] = answer
      
      extra_data = {'worker_id': worker_id, self.main_question: answer, 'accept_time': accept_time, 'submit_time': submit_time}
      if len(self.extra_questions) > 0:
        for q in self.extra_questions:
          extra_data[q] = answer_json[q]
      hit_group_info[hit_id].append(extra_data)
      
      answers_summary[answer] = answers_summary.get(answer, 0) + 1

      self.output['total_assignments'] = self.output.get('total_assignments', 0) + 1
      if assignment_status == 'Approved':
        self.output['total_assignments_approved'] = self.output.get('total_assignments_approved', 0) + 1
      elif assignment_status == 'Rejected':
        self.output['total_assignments_rejected'] = self.output.get('total_assignments_rejected', 0) + 1
    
    self.output['answers_summary'] = answers_summary
    self.hit_group = hit_group
    self.hit_group_info = hit_group_info
    

  def _analyze_hits(self):
    #self.output['total_hits'] = self.db.queryValue("SELECT COUNT(DISTINCT hit_id) FROM {mturk_task_hit} WHERE task_id=?", self.task_id)
    self.output['total_hits'] = len(self.hit_group_info)
    if not self.output['total_hits']:
      self.printMessage('Total HITs is 0, please load HITs first before computing results.')
      raise DCommandExecutionException("Total HITs is 0, please load HITs first before computing results.")
    
    for hit_id, hit_data in self.hit_group_info.items(): 
      json_data = self.db.queryValue("SELECT data FROM {mturk_hit} WHERE hit_id=?", hit_id)
      if json_data is not None:
        data = json.loads(json_data)
        if not isinstance(data, dict):
          data = {}
      else:
        data = {}
      
      data['summary'] = {}  
      data['summary']['total'] = len(hit_data)
      time_spent = [i['submit_time'] - i['accept_time'] for i in hit_data]
      data['summary']['average_time'] = float(sum(time_spent)) / float(len(hit_data))
      data['summary']['median_time'] = sorted(time_spent)[len(time_spent) // 2]
      data['summary'][self.main_question] = '|'.join([i[self.main_question] for i in hit_data])
      if len(self.extra_questions) > 0:
        for q in self.extra_questions:
          data['summary'][q] = '|'.join([i[q] for i in hit_data])
      
      self.db.update("UPDATE {mturk_hit} SET data=? WHERE hit_id=?", json.dumps(data), hit_id)


  def _analyze_assignments(self):
    rows = self.db.queryArray("SELECT MIN(h.max_assignments) m1, MAX(h.max_assignments) m2, SUM(h.max_assignments) m3\
      FROM {mturk_task_hit} t INNER JOIN {mturk_hit} h ON t.hit_id=h.hit_id AND task_id=?", self.task_id)
    m1, m2, m3 = rows[0]
    if m1 == m2:
      self.output['max_assignments'] = int(m1)  # min and max is the same. otherwise 'max_assignments' is unset, and you can't compute fleiss.
    self.output['total_max_assignments'] = m3.intValue()


  def _analyze_workers(self):
    self.output['total_workers'] = self.db.queryValue("SELECT COUNT(DISTINCT a.worker_id) \
        FROM {mturk_task_hit} t INNER JOIN {mturk_assignment} a ON t.hit_id=a.hit_id AND task_id=?", self.task_id)
    # save worker info to output rather than {mturk_worker} because the later is site-wide.
    # to aggregate results to {mturk_worker}, we could write an aggregation logic to take data from DataAnalyze and GAL and then write into {mturk_worker}
    workers = self.db.queryArray("SELECT worker_id, AVG(submit_time - accept_time), COUNT(*) \
        FROM {mturk_task_hit} t INNER JOIN {mturk_assignment} a ON t.hit_id=a.hit_id AND task_id=? GROUP BY worker_id", self.task_id)
    self.output['workers'] = [(worker_id, avg_time.floatValue(), total) for worker_id, avg_time, total in workers]
    

  def _analyze_fleiss(self): 
    if 'max_assignments' in self.output:
      # prepare the matrix
      hit_group = self.hit_group
      hit_group_answers = defaultdict(dict)  # hit_group_answers[hit_id] = {answer: # of answers}
      hit_group_matrix = []  # this is a matrix with rows as hits, and columns as distinct answers, cell values as the number of HITs.
      all_answers = self.output['answers_summary'].keys()
      for hit_id, hit_assignments in hit_group.items():
        # hit_assignments => {worker_id: answer}
        answers = hit_assignments.values()
        answers_group = defaultdict(int)
        for a in answers:
          answers_group[a] += 1
        hit_group_answers[hit_id] = answers_group
        row = []
        for a in all_answers:
          row.append(answers_group[a])
        hit_group_matrix.append(row)
      
      # computing fleiss
      n = self.output['max_assignments']
      mat = [row for row in hit_group_matrix if sum(row) == n]
      self.output['fleiss_valid_hits'] = len(mat)
      if self.output['fleiss_valid_hits'] > 0:
        self.output['fleiss'] = computeKappa(mat)
      else:
        self.output['fleiss'] = None
    else:
      self.output['fleiss'] = None


  def execute(self):
    self.output = {}
    self.db = self.getDrupalSite().getDatabase()

    self._prepare_first()
    self._analyze_hits()
    self._analyze_assignments()
    self._analyze_workers()
    self._analyze_fleiss()

    self.db.close()
    #pprint(self.output)
    self.printMessage('Successfully compute summary.')


  def initialize(self, user_id, task_id, input = {}):
    super(DataAnalysis, self).initialize(user_id, task_id, input)

    # the main question used to compute fleiss/GAL, etc
    self.main_question = input['main_question']
    assert self.main_question is not None
    
    # extra question to be processed.
    self.extra_questions = input.get('extra_questions', [])

#    # how much greater the majority results has to be in order to return a definite results
#    self.majority_advantage = input['analysis_results_majority_advantage']
#    self.majority_advantage = int(self.majority_advantage) if self.majority_advantage is not None else 1
#
#    # if a HIT gets less than this minimum assignments, we'll just skip it.
#    self.minimum_requirement = input['analysis_results_minimum_requirement']
#    self.minimum_requirement = int(self.minimum_requirement) if self.minimum_requirement is not None else 0
#
#    self.ignored_answers = input['analysis_results_ignored_answers']
#    self.ignored_answers = self.ignored_answers.split(',') if self.ignored_answers is not None else []

#    self.results = {}


#  def old_execute(self):
#    self.results['total_hits'] = self.db.queryValue("SELECT COUNT(hit_id) FROM {mturk_task_hit} WHERE task_id=?", self.task_id)
#    if not self.results['total_hits']:
#      self.printMessage('Total HITs is 0, please load HITs first before computing results.')
#      raise DCommandExecutionException("Total HITs is 0, please load HITs first before computing results.")
#
#
#    row = self.db.query("SELECT MIN(h.max_assignments) m1, MAX(h.max_assignments) m2, SUM(h.max_assignments) m3\
#      FROM {mturk_task_hit} t INNER JOIN {mturk_hit} h ON t.hit_id=h.hit_id WHERE task_id=?", self.task_id)
#    row = row[0]
#    m1, m2, m3 = row['m1'], row['m2'], row['m3']
#    if m1 == m2:
#      self.results['max_assignments'] = int(m1)  # min and max is the same. otherwise 'max_assignments' is unset, and you can't compute fleiss.
#    # here m3 is java.BigDecimal and is not recognizable by Jython.
#    self.results['total_max_assignments'] = m3.intValue()
#
#
#    workers_set = set([])
#    answers_summary = {}  # number of answers for each answer (needs to specify the 'main question' setting)
#    assignments = self.db.query("SELECT t.hit_id, a.worker_id, a.assignment_status, a.answer_json\
#        FROM {mturk_task_hit} t INNER JOIN {mturk_assignment} a ON t.hit_id=a.hit_id\
#        WHERE task_id=?", self.task_id)
#    if not len(assignments):
#      self.printMessage('Total assignments is 0, please load assignments first before computing results.')
#      raise DCommandExecutionException("Total assignments is 0, please load assignments first before computing results.")
#
#
#    # hit_group[hit_id] = {worker_id: answer}
#    hit_group = defaultdict(dict)  # group all assignments according to hit_id. if not assignments, then no hit_id as key.
#    for row in assignments:
#      hit_id, worker_id, assignment_status, answer_json = row['hit_id'], row['worker_id'], row['assignment_status'], row['answer_json']
#      answer = self._get_answer(answer_json)
#      hit_group[hit_id][worker_id] = answer
#      workers_set.add(worker_id)
#      answers_summary[answer] = answers_summary.get(answer, 0) + 1
#
#      self.results['total_assignments'] = self.results.get('total_assignments', 0) + 1
#      if assignment_status == 'Approved':
#        self.results['total_assignments_approved'] = self.results.get('total_assignments_approved', 0) + 1
#      elif assignment_status == 'Rejected':
#        self.results['total_assignments_rejected'] = self.results.get('total_assignments_rejected', 0) + 1
#
#
#    self.results['answers_summary'] = answers_summary
#    self.results['total_workers'] = len(workers_set)
#
#
#    hit_group_answers = defaultdict(dict)  # hit_group_answers[hit_id] = {answer: # of answers}
#    hit_group_matrix = []  # this is a matrix with rows as hits, and columns as distinct answers, cell values as the number of HITs.
#    all_answers = self.results['answers_summary'].keys()
#    for hit_id, hit_assignments in hit_group.items():
#      # hit_assignments => {worker_id: answer}
#      answers = hit_assignments.values()
#      answers_group = defaultdict(int)
#      for a in answers:
#        answers_group[a] += 1
#      hit_group_answers[hit_id] = answers_group
#      row = []
#      for a in all_answers:
#        row.append(answers_group[a])
#      hit_group_matrix.append(row)
#      #pprint(hit_group)
#    #pprint(hit_group_answers)
#    #pprint(hit_group_matrix)
#
#
#    # calculate fleiss
#    if 'max_assignments' in self.results:
#      n = self.results['max_assignments']
#      mat = [row for row in hit_group_matrix if sum(row) == n]
#      self.results['fleiss_valid_hits'] = len(mat)
#      if self.results['fleiss_valid_hits'] > 0:
#        self.results['fleiss'] = computeKappa(mat)
#      else:
#        self.results['fleiss'] = 'No finished HITs yet.'
#    else:
#      self.results['fleiss'] = 'All HITs needs to have the same maximum assignments in order to calculate Fleiss Kappa'
#      self.results['fleiss_valid_hits'] = 0
#
#
#    # calculate final results from noisy input
##    self.results['results_majority'] = []
##    rows = self.db.query("SELECT h.hit_id, h.annotation_id FROM {mturk_hit} h INNER JOIN {mturk_task_hit} t ON h.hit_id=t.hit_id WHERE t.task_id=?", self.task_id)
##    for row in rows:
##      hit_id, annotation_id = row['hit_id'], row['annotation_id']
##      combined_answer = ','.join(hit_group[hit_id].values())
##      answers_group = hit_group_answers[hit_id]
##
##      # remove ignored answers
##      for a in self.ignored_answers:
##        if a in answers_group:
##          del(answers_group[a])
##
##      if sum(answers_group.values()) < self.minimum_requirement or sum(answers_group.values()) == 0:
##        final_answer = None
##      else:
##        sorted_answers = answers_group.items() # flatten the dict into a list of tuples
##        sorted_answers.sort(key=lambda r: r[1], reverse=True)
##        diff = sorted_answers[0][1] - sorted_answers[1][1] if len(sorted_answers) > 1 else sorted_answers[0][1]
##        if diff >= self.majority_advantage:
##          final_answer = sorted_answers[0][0]
##        else:
##          final_answer = None
##      self.results['results_majority'].append((hit_id, annotation_id, combined_answer, final_answer))
#
#    #self._save_results()
#    #pprint(self.results)
#    #return Status.SUCCESS, 'successfully computed results.', None
#    self.printMessage('Successfully compute summary.')


  #obsolete
  def _get_answer(self, answer_json):
    answer_json = json.loads(answer_json)
    if self.main_question is not None and self.main_question in answer_json:
      # here we use the 'main question' setting.
      return answer_json[self.main_question]
    elif self.main_question is None and len(answer_json) > 0:
      # if main_question not set, use the first answer
      return answer_json.values()[0]
    return None


#  def execute(self):
#    self.db = self.getDrupalSite().getDatabase()
#    self.old_execute()
#    self.output = self.results
#    pprint(self.output)
#    self.db.close()



def computeKappa(mat):
  """
  Computes the Fleiss' Kappa value as described in (Fleiss, 1971)
  Copied from http://en.wikibooks.org/wiki/Algorithm_Implementation/Statistics/Fleiss'_kappa#Python
  More info about Fleiss' Kappa: http://en.wikipedia.org/wiki/Fleiss%27_kappa

  Computes the Kappa value
      @param n Number of rating per subjects (number of human raters)
      @param mat Matrix[subjects][categories]
      @return The Kappa value
  """

  DEBUG = False

  def checkEachLineCount(mat):
    """
    Assert that each line has a constant number of ratings
      @param mat The matrix checked
      @return The number of ratings
      @throws AssertionError If lines contain different number of ratings
    """
    n = sum(mat[0])
    assert all(sum(line) == n for line in mat[1:]), "Line count != %d (n value)." % n
    return n


  n = checkEachLineCount(mat)   # PRE : every line count must be equal to n
  N = len(mat)
  k = len(mat[0])

  if DEBUG:
    print n, "raters."
    print N, "subjects."
    print k, "categories."

  # Computing p[]
  p = [0.0] * k
  for j in xrange(k):
    p[j] = 0.0
    for i in xrange(N):
      p[j] += mat[i][j]
    p[j] /= N*n
  if DEBUG: print "p =", p

  # Computing P[]
  P = [0.0] * N
  for i in xrange(N):
    P[i] = 0.0
    for j in xrange(k):
      P[i] += mat[i][j] * mat[i][j]
    P[i] = (P[i] - n) / (n * (n - 1))
  if DEBUG: print "P =", P

  # Computing Pbar
  Pbar = sum(P) / N
  if DEBUG: print "Pbar =", Pbar

  # Computing PbarE
  PbarE = 0.0
  for pj in p:
    PbarE += pj * pj
  if DEBUG: print "PbarE =", PbarE

  kappa = (Pbar - PbarE) / (1 - PbarE)
  if DEBUG: print "kappa =", kappa

  return kappa



########################################################


if __name__ == '__main__':
  sys.registry['python.security.respectJavaAccessibility'] = 'false'

  assert sys.registry['python.security.respectJavaAccessibility'] == 'false', 'Please turn off python.security.respectJavaAccessibility.'
  assert sys.version_info[0] >= 2 and sys.version_info[1] >= 7, 'Please use Jython version 2.7+'
  #pprint(sys.path)
  #pprint(sys.currentWorkingDir)
  app = MTurkApplication()
  app.launchFromShell(sys.argv)
